@@ -10,10 +10,11 @@ module LoweredCodeUtils
 # the VS Code extension integration.
 
 using JuliaInterpreter
-using JuliaInterpreter: SSAValue, SlotNumber, Frame
-using JuliaInterpreter: @lookup, moduleof, pc_expr, step_expr!, is_global_ref, is_quotenode_egal, whichtt,
-                        next_until!, finish_and_return!, get_return, nstatements, codelocation, linetable,
-                        is_return, lookup_return
+using JuliaInterpreter: SSAValue, SlotNumber, Frame, Interpreter, RecursiveInterpreter
+using JuliaInterpreter: codelocation, is_global_ref, is_global_ref_egal, is_quotenode_egal, is_return,
+                        lookup, lookup_return, linetable, moduleof, next_until!, nstatements, pc_expr,
+                        step_expr!, whichtt
+using Compiler: Compiler as CC
 
 include("packagedef.jl")
 
